@@ -1,5 +1,5 @@
 # pool.py - Connection pooling for SQLAlchemy
-# Copyright (C) 2005, 2006, 2007 Michael Bayer mike_mp@zzzcomputing.com
+# Copyright (C) 2005, 2006, 2007, 2008 Michael Bayer mike_mp@zzzcomputing.com
 #
 # This module is part of SQLAlchemy and is released under
 # the MIT License: http://www.opensource.org/licenses/mit-license.php
@@ -16,11 +16,11 @@ regular DB-API connect() methods to be transparently managed by a
 SQLAlchemy connection pool.
 """
 
-import weakref, time
+import weakref, time, threading
 
-from sqlalchemy import exceptions, logging
+from sqlalchemy import exc, log
 from sqlalchemy import queue as Queue
-from sqlalchemy.util import thread, threading, pickle
+from sqlalchemy.util import thread, threading, pickle, as_interface
 
 proxies = {}
 
@@ -32,16 +32,13 @@ def manage(module, **params):
     creating new connection pools for each distinct set of connection
     arguments sent to the decorated module's connect() function.
 
-    Arguments:
+    :param module: a DB-API 2.0 database module
 
-    module
-      A DB-API 2.0 database module.
+    :param poolclass: the class used by the pool module to provide
+      pooling.  Defaults to :class:`QueuePool`.
 
-    poolclass
-      The class used by the pool module to provide pooling.  Defaults
-      to ``QueuePool``.
+    :param \*\*params: will be passed through to *poolclass*
 
-    See the ``Pool`` class for options.
     """
     try:
         return proxies[module]
@@ -59,68 +56,55 @@ def clear_managers():
     proxies.clear()
 
 class Pool(object):
-    """Base class for connection pools.
+    """Abstract base class for connection pools."""
 
-    This is an abstract class, implemented by various subclasses
-    including:
+    def __init__(self, creator, recycle=-1, echo=None, use_threadlocal=False,
+                 reset_on_return=True, listeners=None):
+        """
+        Construct a Pool.
 
-    QueuePool
-      Pools multiple connections using ``Queue.Queue``.
+        :param creator: a callable function that returns a DB-API
+          connection object.  The function will be called with
+          parameters.
 
-    SingletonThreadPool
-      Stores a single connection per execution thread.
+        :param recycle: If set to non -1, number of seconds between
+          connection recycling, which means upon checkout, if this
+          timeout is surpassed the connection will be closed and
+          replaced with a newly opened connection. Defaults to -1.
 
-    NullPool
-      Doesn't do any pooling; opens and closes connections.
+        :param echo: If True, connections being pulled and retrieved
+          from the pool will be logged to the standard output, as well
+          as pool sizing information.  Echoing can also be achieved by
+          enabling logging for the "sqlalchemy.pool"
+          namespace. Defaults to False.
 
-    AssertionPool
-      Stores only one connection, and asserts that only one connection
-      is checked out at a time.
+        :param use_threadlocal: If set to True, repeated calls to
+          :meth:`connect` within the same application thread will be
+          guaranteed to return the same connection object, if one has
+          already been retrieved from the pool and has not been
+          returned yet.  Offers a slight performance advantage at the
+          cost of individual transactions by default.  The
+          :meth:`unique_connection` method is provided to bypass the
+          threadlocal behavior installed into :meth:`connect`.
 
-    The main argument, `creator`, is a callable function that returns
-    a newly connected DB-API connection object.
+        :param reset_on_return: If true, reset the database state of
+          connections returned to the pool.  This is typically a
+          ROLLBACK to release locks and transaction resources.
+          Disable at your own peril.  Defaults to True.
 
-    Options that are understood by Pool are:
+        :param listeners: A list of
+          :class:`~sqlalchemy.interfaces.PoolListener`-like objects or
+          dictionaries of callables that receive events when DB-API
+          connections are created, checked out and checked in to the
+          pool.
 
-    echo
-      If set to True, connections being pulled and retrieved from/to
-      the pool will be logged to the standard output, as well as pool
-      sizing information.  Echoing can also be achieved by enabling
-      logging for the "sqlalchemy.pool" namespace. Defaults to False.
-
-    use_threadlocal
-      If set to True, repeated calls to ``connect()`` within the same
-      application thread will be guaranteed to return the same
-      connection object, if one has already been retrieved from the
-      pool and has not been returned yet. This allows code to retrieve
-      a connection from the pool, and then while still holding on to
-      that connection, to call other functions which also ask the pool
-      for a connection of the same arguments; those functions will act
-      upon the same connection that the calling method is using.
-      Defaults to True.
-
-    recycle
-      If set to non -1, a number of seconds between connection
-      recycling, which means upon checkout, if this timeout is
-      surpassed the connection will be closed and replaced with a
-      newly opened connection. Defaults to -1.
-
-    listeners
-      A list of ``PoolListener``-like objects that receive events when
-      DB-API connections are created, checked out and checked in to
-      the pool.
-    """
-
-    def __init__(self, creator, recycle=-1, echo=None, use_threadlocal=True,
-                 listeners=None):
-        self.logger = logging.instance_logger(self, echoflag=echo)
-        # the WeakValueDictionary works more nicely than a regular dict
-        # of weakrefs.  the latter can pile up dead reference objects which don't
-        # get cleaned out.  WVD adds from 1-6 method calls to a checkout operation.
-        self._threadconns = weakref.WeakValueDictionary()
+        """
+        self.logger = log.instance_logger(self, echoflag=echo)
+        self._threadconns = threading.local()
         self._creator = creator
         self._recycle = recycle
         self._use_threadlocal = use_threadlocal
+        self._reset_on_return = reset_on_return
         self.echo = echo
         self.listeners = []
         self._on_connect = []
@@ -158,15 +142,19 @@ class Pool(object):
             return _ConnectionFairy(self).checkout()
 
         try:
-            return self._threadconns[thread.get_ident()].checkout()
-        except KeyError:
-            agent = _ConnectionFairy(self)
-            self._threadconns[thread.get_ident()] = agent
-            return agent.checkout()
+            rec = self._threadconns.current()
+            if rec:
+                return rec.checkout()
+        except AttributeError:
+            pass
+
+        agent = _ConnectionFairy(self)
+        self._threadconns.current = weakref.ref(agent)
+        return agent.checkout()
 
     def return_conn(self, record):
-        if self._use_threadlocal and thread.get_ident() in self._threadconns:
-            del self._threadconns[thread.get_ident()]
+        if self._use_threadlocal and hasattr(self._threadconns, "current"):
+            del self._threadconns.current
         self.do_return_conn(record)
 
     def get(self):
@@ -182,7 +170,16 @@ class Pool(object):
         raise NotImplementedError()
 
     def add_listener(self, listener):
-        """Add a ``PoolListener``-like object to this pool."""
+        """Add a ``PoolListener``-like object to this pool.
+
+        ``listener`` may be an object that implements some or all of
+        PoolListener, or a dictionary of callables containing implementations
+        of some or all of the named methods in PoolListener.
+
+        """
+
+        listener = as_interface(
+            listener, methods=('connect', 'checkout', 'checkin'))
 
         self.listeners.append(listener)
         if hasattr(listener, 'connect'):
@@ -207,15 +204,23 @@ class _ConnectionRecord(object):
     def close(self):
         if self.connection is not None:
             if self.__pool._should_log_info:
-                self.__pool.log("Closing connection %s" % repr(self.connection))
-            self.connection.close()
+                self.__pool.log("Closing connection %r" % self.connection)
+            try:
+                self.connection.close()
+            except (SystemExit, KeyboardInterrupt):
+                raise
+            except:
+                if self.__pool._should_log_info:
+                    self.__pool.log("Exception closing connection %r" %
+                                    self.connection)
 
     def invalidate(self, e=None):
         if self.__pool._should_log_info:
             if e is not None:
-                self.__pool.log("Invalidate connection %s (reason: %s:%s)" % (repr(self.connection), e.__class__.__name__, str(e)))
+                self.__pool.log("Invalidate connection %r (reason: %s:%s)" %
+                                (self.connection, e.__class__.__name__, e))
             else:
-                self.__pool.log("Invalidate connection %s" % repr(self.connection))
+                self.__pool.log("Invalidate connection %r" % self.connection)
         self.__close()
         self.connection = None
 
@@ -228,7 +233,8 @@ class _ConnectionRecord(object):
                     l.connect(self.connection, self)
         elif (self.__pool._recycle > -1 and time.time() - self.starttime > self.__pool._recycle):
             if self.__pool._should_log_info:
-                self.__pool.log("Connection %s exceeded timeout; recycling" % repr(self.connection))
+                self.__pool.log("Connection %r exceeded timeout; recycling" %
+                                self.connection)
             self.__close()
             self.connection = self.__connect()
             self.info.clear()
@@ -240,11 +246,12 @@ class _ConnectionRecord(object):
     def __close(self):
         try:
             if self.__pool._should_log_info:
-                self.__pool.log("Closing connection %s" % (repr(self.connection)))
+                self.__pool.log("Closing connection %r" % self.connection)
             self.connection.close()
         except Exception, e:
             if self.__pool._should_log_info:
-                self.__pool.log("Connection %s threw an error on close: %s" % (repr(self.connection), str(e)))
+                self.__pool.log("Connection %r threw an error on close: %s" %
+                                (self.connection, e))
             if isinstance(e, (SystemExit, KeyboardInterrupt)):
                 raise
 
@@ -253,22 +260,21 @@ class _ConnectionRecord(object):
             self.starttime = time.time()
             connection = self.__pool._creator()
             if self.__pool._should_log_info:
-                self.__pool.log("Created new connection %s" % repr(connection))
+                self.__pool.log("Created new connection %r" % connection)
             return connection
         except Exception, e:
             if self.__pool._should_log_info:
-                self.__pool.log("Error on connect(): %s" % (str(e)))
+                self.__pool.log("Error on connect(): %s" % e)
             raise
 
-    properties = property(lambda self: self.info,
-                          doc="A synonym for .info, will be removed in 0.5.")
 
 def _finalize_fairy(connection, connection_record, pool, ref=None):
     if ref is not None and connection_record.backref is not ref:
         return
     if connection is not None:
         try:
-            connection.rollback()
+            if pool._reset_on_return:
+                connection.rollback()
             # Immediately close detached instances
             if connection_record is None:
                 connection.close()
@@ -280,7 +286,7 @@ def _finalize_fairy(connection, connection_record, pool, ref=None):
     if connection_record is not None:
         connection_record.backref = None
         if pool._should_log_info:
-            pool.log("Connection %s being returned to pool" % repr(connection))
+            pool.log("Connection %r being returned to pool" % connection)
         if pool._on_checkin:
             for l in pool._on_checkin:
                 l.checkin(connection, connection_record)
@@ -289,6 +295,8 @@ def _finalize_fairy(connection, connection_record, pool, ref=None):
 class _ConnectionFairy(object):
     """Proxies a DB-API connection and provides return-on-dereference support."""
 
+    __slots__ = '_pool', '__counter', 'connection', '_connection_record', '__weakref__', '_detached_info'
+    
     def __init__(self, pool):
         self._pool = pool
         self.__counter = 0
@@ -301,27 +309,31 @@ class _ConnectionFairy(object):
             self._connection_record = None
             raise
         if self._pool._should_log_info:
-            self._pool.log("Connection %s checked out from pool" % repr(self.connection))
+            self._pool.log("Connection %r checked out from pool" %
+                           self.connection)
 
-    _logger = property(lambda self: self._pool.logger)
+    @property
+    def _logger(self):
+        return self._pool.logger
 
-    is_valid = property(lambda self:self.connection is not None)
+    @property
+    def is_valid(self):
+        return self.connection is not None
 
-    def _get_info(self):
+    @property
+    def info(self):
         """An info collection unique to this DB-API connection."""
 
         try:
             return self._connection_record.info
         except AttributeError:
             if self.connection is None:
-                raise exceptions.InvalidRequestError("This connection is closed")
+                raise exc.InvalidRequestError("This connection is closed")
             try:
                 return self._detached_info
             except AttributeError:
                 self._detached_info = value = {}
                 return value
-    info = property(_get_info)
-    properties = property(_get_info)
 
     def invalidate(self, e=None):
         """Mark this connection as invalidated.
@@ -331,7 +343,7 @@ class _ConnectionFairy(object):
         """
 
         if self.connection is None:
-            raise exceptions.InvalidRequestError("This connection is closed")
+            raise exc.InvalidRequestError("This connection is closed")
         if self._connection_record is not None:
             self._connection_record.invalidate(e=e)
         self.connection = None
@@ -350,8 +362,8 @@ class _ConnectionFairy(object):
 
     def checkout(self):
         if self.connection is None:
-            raise exceptions.InvalidRequestError("This connection is closed")
-        self.__counter +=1
+            raise exc.InvalidRequestError("This connection is closed")
+        self.__counter += 1
 
         if not self._pool._on_checkout or self.__counter != 1:
             return self
@@ -363,10 +375,10 @@ class _ConnectionFairy(object):
                 for l in self._pool._on_checkout:
                     l.checkout(self.connection, self._connection_record, self)
                 return self
-            except exceptions.DisconnectionError, e:
+            except exc.DisconnectionError, e:
                 if self._pool._should_log_info:
                     self._pool.log(
-                    "Disconnection detected on checkout: %s" % (str(e)))
+                    "Disconnection detected on checkout: %s" % e)
                 self._connection_record.invalidate(e)
                 self.connection = self._connection_record.get_connection()
                 attempts -= 1
@@ -374,7 +386,7 @@ class _ConnectionFairy(object):
         if self._pool._should_log_info:
             self._pool.log("Reconnection attempts exhausted on checkout")
         self.invalidate()
-        raise exceptions.InvalidRequestError("This connection is closed")
+        raise exc.InvalidRequestError("This connection is closed")
 
     def detach(self):
         """Separate this connection from its Pool.
@@ -398,7 +410,7 @@ class _ConnectionFairy(object):
             self._connection_record = None
 
     def close(self):
-        self.__counter -=1
+        self.__counter -= 1
         if self.__counter == 0:
             self._close()
 
@@ -408,10 +420,13 @@ class _ConnectionFairy(object):
         self._connection_record = None
 
 class _CursorFairy(object):
+    __slots__ = '__parent', 'cursor', 'execute'
+
     def __init__(self, parent, cursor):
         self.__parent = parent
         self.cursor = cursor
-
+        self.execute = cursor.execute
+        
     def invalidate(self, e=None):
         self.__parent.invalidate(e=e)
 
@@ -419,7 +434,12 @@ class _CursorFairy(object):
         try:
             self.cursor.close()
         except Exception, e:
-            self.__parent._logger.warn("Error closing cursor: " + str(e))
+            try:
+                ex_text = str(e)
+            except TypeError:
+                ex_text = repr(e)
+            self.__parent._logger.warn("Error closing cursor: " + ex_text)
+
             if isinstance(e, (SystemExit, KeyboardInterrupt)):
                 raise
 
@@ -436,105 +456,142 @@ class SingletonThreadPool(Pool):
     default, and also requires a singleton connection if a :memory: database
     is being used.
 
-    Options are the same as those of Pool, as well as:
+    Options are the same as those of :class:`Pool`, as well as:
 
-    pool_size: 5
-      The number of threads in which to maintain connections at once.
+    :param pool_size: The number of threads in which to maintain connections 
+        at once.  Defaults to five.
+      
     """
 
     def __init__(self, creator, pool_size=5, **params):
         params['use_threadlocal'] = True
         Pool.__init__(self, creator, **params)
-        self._conns = {}
+        self._conn = threading.local()
+        self._all_conns = set()
         self.size = pool_size
 
     def recreate(self):
         self.log("Pool recreating")
-        return SingletonThreadPool(self._creator, pool_size=self.size, recycle=self._recycle, echo=self._should_log_info, use_threadlocal=self._use_threadlocal, listeners=self.listeners)
+        return SingletonThreadPool(self._creator, 
+            pool_size=self.size, 
+            recycle=self._recycle, 
+            echo=self._should_log_info, 
+            use_threadlocal=self._use_threadlocal, 
+            listeners=self.listeners)
 
     def dispose(self):
-        """Dispose of this pool.
+        """Dispose of this pool."""
 
-        this method leaves the possibility of checked-out connections
-        remaining opened, so it is advised to not reuse the pool once
-        dispose() is called, and to instead use a new pool constructed
-        by the recreate() method.
-        """
-
-        for key, conn in self._conns.items():
+        for conn in self._all_conns:
             try:
                 conn.close()
             except (SystemExit, KeyboardInterrupt):
                 raise
             except:
-                # sqlite won't even let you close a conn from a thread
+                # pysqlite won't even let you close a conn from a thread
                 # that didn't create it
                 pass
-            del self._conns[key]
-
+        
+        self._all_conns.clear()
+            
     def dispose_local(self):
-        try:
-            del self._conns[thread.get_ident()]
-        except KeyError:
-            pass
+        if hasattr(self._conn, 'current'):
+            conn = self._conn.current()
+            self._all_conns.discard(conn)
+            del self._conn.current
 
     def cleanup(self):
-        for key in self._conns.keys():
-            try:
-                del self._conns[key]
-            except KeyError:
-                pass
-            if len(self._conns) <= self.size:
+        for conn in list(self._all_conns):
+            self._all_conns.discard(conn)
+            if len(self._all_conns) <= self.size:
                 return
 
     def status(self):
-        return "SingletonThreadPool id:%d thread:%d size: %d" % (id(self), thread.get_ident(), len(self._conns))
+        return "SingletonThreadPool id:%d size: %d" % (id(self), len(self._all_conns))
 
     def do_return_conn(self, conn):
         pass
 
     def do_get(self):
         try:
-            return self._conns[thread.get_ident()]
-        except KeyError:
-            c = self.create_connection()
-            self._conns[thread.get_ident()] = c
-            if len(self._conns) > self.size:
-                self.cleanup()
-            return c
+            c = self._conn.current()
+            if c:
+                return c
+        except AttributeError:
+            pass
+        c = self.create_connection()
+        self._conn.current = weakref.ref(c)
+        self._all_conns.add(c)
+        if len(self._all_conns) > self.size:
+            self.cleanup()
+        return c
 
 class QueuePool(Pool):
-    """A Pool that imposes a limit on the number of open connections.
+    """A Pool that imposes a limit on the number of open connections."""
 
-    Arguments include all those used by the base Pool class, as well
-    as:
+    def __init__(self, creator, pool_size=5, max_overflow=10, timeout=30,
+                 **params):
+        """
+        Construct a QueuePool.
 
-    pool_size
-      The size of the pool to be maintained. This is the largest
-      number of connections that will be kept persistently in the
-      pool. Note that the pool begins with no connections; once this
-      number of connections is requested, that number of connections
-      will remain. Defaults to 5.
+        :param creator: a callable function that returns a DB-API
+          connection object.  The function will be called with
+          parameters.
 
-    max_overflow
-      The maximum overflow size of the pool. When the number of
-      checked-out connections reaches the size set in pool_size,
-      additional connections will be returned up to this limit. When
-      those additional connections are returned to the pool, they are
-      disconnected and discarded. It follows then that the total
-      number of simultaneous connections the pool will allow is
-      pool_size + `max_overflow`, and the total number of "sleeping"
-      connections the pool will allow is pool_size. `max_overflow` can
-      be set to -1 to indicate no overflow limit; no limit will be
-      placed on the total number of concurrent connections. Defaults
-      to 10.
+        :param pool_size: The size of the pool to be maintained. This
+          is the largest number of connections that will be kept
+          persistently in the pool. Note that the pool begins with no
+          connections; once this number of connections is requested,
+          that number of connections will remain. Defaults to 5.
 
-    timeout
-      The number of seconds to wait before giving up on returning a
-      connection. Defaults to 30.
-    """
+        :param max_overflow: The maximum overflow size of the
+          pool. When the number of checked-out connections reaches the
+          size set in pool_size, additional connections will be
+          returned up to this limit. When those additional connections
+          are returned to the pool, they are disconnected and
+          discarded. It follows then that the total number of
+          simultaneous connections the pool will allow is pool_size +
+          `max_overflow`, and the total number of "sleeping"
+          connections the pool will allow is pool_size. `max_overflow`
+          can be set to -1 to indicate no overflow limit; no limit
+          will be placed on the total number of concurrent
+          connections. Defaults to 10.
 
-    def __init__(self, creator, pool_size = 5, max_overflow = 10, timeout=30, **params):
+        :param timeout: The number of seconds to wait before giving up
+          on returning a connection. Defaults to 30.
+
+        :param recycle: If set to non -1, number of seconds between
+          connection recycling, which means upon checkout, if this
+          timeout is surpassed the connection will be closed and
+          replaced with a newly opened connection. Defaults to -1.
+
+        :param echo: If True, connections being pulled and retrieved
+          from the pool will be logged to the standard output, as well
+          as pool sizing information.  Echoing can also be achieved by
+          enabling logging for the "sqlalchemy.pool"
+          namespace. Defaults to False.
+
+        :param use_threadlocal: If set to True, repeated calls to
+          :meth:`connect` within the same application thread will be
+          guaranteed to return the same connection object, if one has
+          already been retrieved from the pool and has not been
+          returned yet.  Offers a slight performance advantage at the
+          cost of individual transactions by default.  The
+          :meth:`unique_connection` method is provided to bypass the
+          threadlocal behavior installed into :meth:`connect`.
+
+        :param reset_on_return: If true, reset the database state of
+          connections returned to the pool.  This is typically a
+          ROLLBACK to release locks and transaction resources.
+          Disable at your own peril.  Defaults to True.
+
+        :param listeners: A list of
+          :class:`~sqlalchemy.interfaces.PoolListener`-like objects or
+          dictionaries of callables that receive events when DB-API
+          connections are created, checked out and checked in to the
+          pool.
+
+        """
         Pool.__init__(self, creator, **params)
         self._pool = Queue.Queue(pool_size)
         self._overflow = 0 - pool_size
@@ -568,7 +625,7 @@ class QueuePool(Pool):
                 if not wait:
                     return self.do_get()
                 else:
-                    raise exceptions.TimeoutError("QueuePool limit of size %d overflow %d reached, connection timed out, timeout %d" % (self.size(), self.overflow(), self._timeout))
+                    raise exc.TimeoutError("QueuePool limit of size %d overflow %d reached, connection timed out, timeout %d" % (self.size(), self.overflow(), self._timeout))
 
             if self._overflow_lock is not None:
                 self._overflow_lock.acquire()
@@ -619,24 +676,73 @@ class NullPool(Pool):
 
     Instead it literally opens and closes the underlying DB-API connection
     per each connection open/close.
+
+    Reconnect-related functions such as ``recycle`` and connection
+    invalidation are not supported by this Pool implementation, since
+    no connections are held persistently.
+
     """
 
     def status(self):
         return "NullPool"
 
     def do_return_conn(self, conn):
-       conn.close()
+        conn.close()
 
     def do_return_invalid(self, conn):
-       pass
+        pass
 
     def do_get(self):
         return self.create_connection()
 
+    def recreate(self):
+        self.log("Pool recreating")
+
+        return NullPool(self._creator, 
+            recycle=self._recycle, 
+            echo=self._should_log_info, 
+            use_threadlocal=self._use_threadlocal, 
+            listeners=self.listeners)
+
+    def dispose(self):
+        pass
+        
 class StaticPool(Pool):
-    """A Pool of exactly one connection, used for all requests."""
+    """A Pool of exactly one connection, used for all requests.
+    
+    Reconnect-related functions such as ``recycle`` and connection
+    invalidation (which is also used to support auto-reconnect) are not
+    currently supported by this Pool implementation but may be implemented
+    in a future release.
+    
+    """
 
     def __init__(self, creator, **params):
+        """
+        Construct a StaticPool.
+
+        :param creator: a callable function that returns a DB-API
+          connection object.  The function will be called with
+          parameters.
+
+        :param echo: If True, connections being pulled and retrieved
+          from the pool will be logged to the standard output, as well
+          as pool sizing information.  Echoing can also be achieved by
+          enabling logging for the "sqlalchemy.pool"
+          namespace. Defaults to False.
+
+        :param reset_on_return: If true, reset the database state of
+          connections returned to the pool.  This is typically a
+          ROLLBACK to release locks and transaction resources.
+          Disable at your own peril.  Defaults to True.
+
+        :param listeners: A list of
+          :class:`~sqlalchemy.interfaces.PoolListener`-like objects or
+          dictionaries of callables that receive events when DB-API
+          connections are created, checked out and checked in to the
+          pool.
+
+        """
         Pool.__init__(self, creator, **params)
         self._conn = creator()
         self.connection = _ConnectionRecord(self)
@@ -644,6 +750,10 @@ class StaticPool(Pool):
     def status(self):
         return "StaticPool"
 
+    def dispose(self):
+        self._conn.close()
+        self._conn = None
+
     def create_connection(self):
         return self._conn
 
@@ -663,11 +773,51 @@ class AssertionPool(Pool):
     This will raise an exception if more than one connection is checked out
     at a time.  Useful for debugging code that is using more connections
     than desired.
+
     """
 
     ## TODO: modify this to handle an arbitrary connection count.
 
     def __init__(self, creator, **params):
+        """
+        Construct an AssertionPool.
+
+        :param creator: a callable function that returns a DB-API
+          connection object.  The function will be called with
+          parameters.
+
+        :param recycle: If set to non -1, number of seconds between
+          connection recycling, which means upon checkout, if this
+          timeout is surpassed the connection will be closed and
+          replaced with a newly opened connection. Defaults to -1.
+
+        :param echo: If True, connections being pulled and retrieved
+          from the pool will be logged to the standard output, as well
+          as pool sizing information.  Echoing can also be achieved by
+          enabling logging for the "sqlalchemy.pool"
+          namespace. Defaults to False.
+
+        :param use_threadlocal: If set to True, repeated calls to
+          :meth:`connect` within the same application thread will be
+          guaranteed to return the same connection object, if one has
+          already been retrieved from the pool and has not been
+          returned yet.  Offers a slight performance advantage at the
+          cost of individual transactions by default.  The
+          :meth:`unique_connection` method is provided to bypass the
+          threadlocal behavior installed into :meth:`connect`.
+
+        :param reset_on_return: If true, reset the database state of
+          connections returned to the pool.  This is typically a
+          ROLLBACK to release locks and transaction resources.
+          Disable at your own peril.  Defaults to True.
+
+        :param listeners: A list of
+          :class:`~sqlalchemy.interfaces.PoolListener`-like objects or
+          dictionaries of callables that receive events when DB-API
+          connections are created, checked out and checked in to the
+          pool.
+
+        """
         Pool.__init__(self, creator, **params)
         self.connection = _ConnectionRecord(self)
         self._conn = self.connection
@@ -676,14 +826,14 @@ class AssertionPool(Pool):
         return "AssertionPool"
 
     def create_connection(self):
-        raise "Invalid"
+        raise AssertionError("Invalid")
 
     def do_return_conn(self, conn):
         assert conn is self._conn and self.connection is None
         self.connection = conn
 
     def do_return_invalid(self, conn):
-        raise "Invalid"
+        raise AssertionError("Invalid")
 
     def do_get(self):
         assert self.connection is not None
@@ -715,7 +865,8 @@ class _DBProxy(object):
         self.params = params
         self.poolclass = poolclass
         self.pools = {}
-
+        self._create_pool_mutex = threading.Lock()
+        
     def close(self):
         for key in self.pools.keys():
             del self.pools[key]
@@ -731,10 +882,17 @@ class _DBProxy(object):
         try:
             return self.pools[key]
         except KeyError:
-            pool = self.poolclass(lambda: self.module.connect(*args, **params), **self.params)
-            self.pools[key] = pool
-            return pool
-
+            self._create_pool_mutex.acquire()
+            try:
+                if key not in self.pools:
+                    pool = self.poolclass(lambda: self.module.connect(*args, **params), **self.params)
+                    self.pools[key] = pool
+                    return pool
+                else:
+                    return self.pools[key]
+            finally:
+                self._create_pool_mutex.release()
+                
     def connect(self, *args, **params):
         """Activate a connection to the database.
 
